%% tclp__clp_types2.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__clp_types2.pl,v 1.3 2003/07/29 15:40:38 manu Exp $

:- module(tclp__clp_types2, [':<'/2,
			     fresh/1,
			     fresh_term/1,
			     init/0,
			     types_reinit/0,
			     getbounds/3,
			     check_constraint/1,
			     term_type_up/2,
			     term_type_down/2,
			     cyclic/1, cyclic_up/1, cyclic_down/1,
			     type_info/2,
			     get_greater_vars/2,
			     get_lesser_vars/2,
			     try_generalize_var/2,
			     ordered_vars/2,
			     tclp__vGLB/2,
			     tclp__vLUB/2
                            ]).

:- use_module(tclp(tclp__def)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__type_order)).
:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__variables_set_assoc)).
:- use_module(library(chr)).
:- use_module(library(ordsets)).
:- use_module(library(lists)).

%% Verbose modes: debug_clp_types2

handler type2.

constraints %% delaying some contraints
            delay/2,
            %% upper bound of a variable
            hibound/2, canonic_hibound/2,
            %% lower bound of a variable
            lobound/2, canonic_lobound/2,
            %% sets of variables over/under a given variable
            hivars/2, lovars/2,
            %% original and introduced variables
            is_original/1, glb_variable/2, lub_variable/2,
            %% updating variables bounds and up/lo sets
            update_v_lo/2, update_v_hi/2.

%% canonic_xxbound(X,'$none') stands for:
%% X has no current canonical (lo/hi) bound

operator(100,xfx,[':<']).

%%%%%%%%%
%% Test if some variable is a type variable
is_type_variable(X) :- chr:find_constraint(X,canonic_lobound(X,_),_),!.

%%%%%%%%%
%% Dealing with bounds, variables and unification

%%%%%
%% If there is no canonic_xxbound, use any bound as a canonic_xxbound
def_canonic_hi @
  canonic_hibound(X,'$none'), hibound(X,B) <=> canonic_hibound(X,B).
def_canonic_lo @
  canonic_lobound(X,'$none'), lobound(X,B) <=> canonic_lobound(X,B).

%%%%%
%% Delete immediate consequences
immediate1 @
  canonic_hibound(X,CB) \ hibound(X,B)
        <=> ordered_types(CB,B) | true.
immediate2 @
  canonic_lobound(X,CB) \ lobound(X,B)
        <=> ordered_types(B,CB) | true.

%%%%%%%%%
%% Decomposition of constraints

%%%%%
%% Rules
dec_canonical @
  canonic_lobound(X,LB), canonic_hibound(X,HB) ==> try_break(LB,HB).
dec_canonical_hi @
  canonic_hibound(X,HB), lobound(X,LB) ==> try_break(LB,HB).
dec_canonical_lo @
  canonic_lobound(X,LB), hibound(X,HB) ==> try_break(LB,HB).
dec @
  hibound(X,HB), lobound(X,LB) ==> try_break(LB,HB).

%%%%%
%% breaking a constraint (deals with $none)
try_break('$none',_) :- !.
try_break(_,'$none') :- !.
try_break(Tau,Tau2) :- tclp__break(Tau,Tau2).

canonic_lobound(X,_), canonic_lobound(Y,_) \ delay(X,Y) <=> X:<Y.
canonic_lobound(X,_), delay(X,Y) <=> nonvar(Y) | X:<Y.
canonic_lobound(Y,_), delay(X,Y) <=> nonvar(X) | X:<Y.
delay(X,Y) <=> nonvar(X),  nonvar(Y) | X:<Y.

%%%%%%%%%
%% posting Tau :< Tau'
X :< Y :- var(X), \+ is_type_variable(X),!, delay(X,Y).
X :< Y :- var(Y), \+ is_type_variable(Y),!, delay(X,Y).
X :< Y :- var(X),var(Y), !, put_var_leq(X,Y).
X :< Tau :- var(X),nonvar(Tau), !, put_hibound(X,Tau).
Tau :< X :- var(X),nonvar(Tau), !, put_lobound(X,Tau).
Tau :< Tau2 :- nonvar(Tau), nonvar(Tau2), !, put_break(Tau,Tau2).

%%%%%%%%%
%% inequalities between variables
put_var_leq(X,Y) :- X==Y,!.
put_var_leq(X,Y) :-
        find_constraint(X,hivars(X,L),_),
        member(Z,L),Y==Z,!.
put_var_leq(X,Y) :-
        find_constraint(X,lovars(X,LoVX),_),
        find_constraint(Y,hivars(Y,UpVY),_),
        update_lo([Y|UpVY],X),
        update_hi([X|LoVX],Y).

%%%%%%%%%
%% updating variable data

%%%%%
%% Lo bounds

%% update_lo(L,X) propagate for each element Y of L that X is smaller than Y
%% (that is propagate to Y the lower bounds of X and the variables under X)
update_lo([], _).
update_lo([X|Rest],Y) :-
        update_v_lo(X,Y),
        update_lo(Rest,Y).

%% first update bounds
update_lobound @
  update_v_lo(X,Y), lobound(Y,B) ==> lobound(X,B).
%% use also the current canonical bound of Y
update_lobound_canonical
  update_v_lo(X,Y), canonic_lobound(Y,B) ==> lobound(X,B).
%% lastly update variables
update_lovars @
  lovars(Y,LY) \ update_v_lo(X,Y), lovars(X,LX)
        <=> list_to_ord_set(LY,LYS),
            list_to_ord_set(LX,LXS),
            ord_union(LXS,LYS,LXYS),
            ord_add_element(LXYS,Y,NewLX),
            lovars(X,NewLX).

%%%%%
%% Up bounds

%% update_up(L,X) propagate for each element Y of L that X is greater than Y
%% (that is propagate to Y the upper bounds of X and the variables over X)
update_hi([], _).
update_hi([X|Rest],Y) :-
        update_v_hi(X,Y),
        update_hi(Rest,Y).

%% first update bounds
update_hibound @
  update_v_hi(X,Y), hibound(Y,B) ==> hibound(X,B).
%% use also the current canonical bound of Y
update_hibound_canonical
  update_v_hi(X,Y), canonic_hibound(Y,B) ==> hibound(X,B).
%% lastly update variables
update_hivars @
  hivars(Y,UY) \ update_v_hi(X,Y), hivars(X,UX)
        <=> list_to_ord_set(UY,UYS),
            list_to_ord_set(UX,UXS),
            ord_union(UXS,UYS,UXYS),
            ord_add_element(UXYS,Y,NewUX),
            hivars(X,NewUX).


%%%%%%%%%
%% Creating fresh variables

%% adds basic constraints correspinding to a new type variable
core_fresh(X) :-
        canonic_hibound(X,'$none'),
        canonic_lobound(X,'$none'),
        hivars(X,[]),
        lovars(X,[]).

%% add constraints such that X is a fresh original variable
fresh(X) :-
        core_fresh(X), is_original(X), register(X).
%% add constraints such that X is a fresh introduced variable for the glb
%% of Origs
fresh_glb(X,Origs) :-
        core_fresh(X), glb_variable(Origs,X).
%% add constraints such that X is a fresh introduced variable for the lub
%% of Origs
fresh_lub(X,Origs) :-
        core_fresh(X), lub_variable(Origs,X).

%%%%%%%%%
%% Computing new canonical bounds

%% retreive original variables of X
original_lo(X,XLo) :-
	chr:find_constraint(X,lub_variable(Vars,X),_),get_ordset(XLo,Vars)!.
original_lo(X,[X]).

%% ensure that X is declared as a lub of some variables
ensure_lub(X,Origs) :- chr:find_constraint(X,lub_variable(_,X),_),!.
ensure_lub(X,Origs) :- fresh_lub(X,Origs).

%% retrieve the LUB of a list of variables or create it
get_vLUB([X],X).
get_vLUB(Origs,LUB) :-
	associated(Origs,glblub(_,LUB)),
	ensure_lub(LUB, Origs).

%% retreive the LUB
tclp__vLUB(X,Y,Z) :- X==Y,!,X=Z.
%% tclp__vLUB(X,Y,LUB) :- ordered_vars(X,Y),!,LUB=Y.
%% tclp__vLUB(X,Y,LUB) :- ordered_vars(Y,X),!,LUB=X.
tclp__vLUB(X,Y,LUB) :-
	original_lo(X,XLo), original_lo(Y,YLo), ord_union(XLo,YLo,AllLo),
	get_vLUB(AllLo,LUB), X :< LUB, Y :< LUB, !.

%% retreive original variables of X
original_hi(X,XHi) :-
	chr:find_constraint(X,glb_variable(Vars,X),_),get_ordset(XHi,Vars)!.
original_hi(X,[X]).

%% ensure that X is declared as a lub of some variables
ensure_glb(X,Origs) :- chr:find_constraint(X,glb_variable(_,X),_),!.
ensure_glb(X,Origs) :- fresh_glb(X,Origs).

%% retrieve the GLB of a list of variables or create it
get_vGLB([X],X).
get_vGLB(Origs,GLB) :-
	associated(Origs,glblub(GLB,_)),
	ensure_glb(GLB, Origs).

%% retreive the GLB
tclp__vGLB(X,Y,Z) :- X==Y,!,X=Z.
%% tclp__vGLB(X,Y,GLB) :- ordered_vars(X,Y),!,GLB=X.
%% tclp__vGLB(X,Y,GLB) :- ordered_vars(Y,X),!,GLB=Y.
tclp__vGLB(X,Y,GLB) :-
	original_hi(X,XHi), original_hi(Y,YHi), ord_union(XHi,YHi,AllHi),
	get_vGLB(AllHi,GLB), X :< GLB, Y :< GLB, !.

%%%%%%%%%
%% multiple LUB and GLB

%% retrieve original variables of a list of LUBs
all_original_lo([],[]).
all_original_lo([X|Rest],Lows) :-
	original_lo(X,XLo),
	all_original_lo(Rest,OrigRest),
	ord_union(XLo,OrigRest,Lows).

%% retrieve the LUB of a list of variables
tclp__multiple_vLUB(List,LUB) :-
	all_original_lo(List,Origs),
	get_vLUB(Origs,LUB), link(List,[LUB]).

%% retrieve original variables of a list of GLBs
all_original_hi([],[]).
all_original_hi([X|Rest],His) :-
	original_hi(X,XHi),
	all_original_hi(Rest,OrigRest),
	ord_union(XHi,OrigRest,His).

%% retrieve the GLB of a list of variables
tclp__multiple_vGLB(List,GLB) :-
	all_original_hi(List,Origs),
	get_vGLB(Origs,GLB), link([GLB],List).

%% replace_by_originals_lo(Tau,OTau)
%% OTau is the type constructor of Tau with the list of original variable
%% under the urguments of Tau as its arguments
replace_by_originals_lo(Tau,OTau) :-
	Tau =.. [ C | Args ],
	map_originals_lo(Args,OArgs),
	OTau =.. [ C | OArgs ].
map_originals_lo([],[]).
map_originals_lo([X|Rest],[XLo|OtherOrigs]) :-
	original_lo(X,XLo),
	map_originals_lo(Rest,OtherOrigs).

%% replace_by_originals_hi(Tau,OTau)
%% OTau is the type constructor of Tau with the list of original variable
%% under the urguments of Tau as its arguments
replace_by_originals_hi(Tau,OTau) :-
	Tau =.. [ C | Args ],
	map_originals_hi(Args,OArgs),
	OTau =.. [ C | OArgs ].
map_originals_hi([],[]).
map_originals_hi([X|Rest],[XHi|OtherOrigs]) :-
	original_hi(X,XHi),
	map_originals_hi(Rest,OtherOrigs).

%% replace_by_lub(OTau,Tau)
%% OTau is the type constructor of Tau with the list of original variable
%% under the arguments of Tau as its arguments
replace_by_lub(OTau,Tau) :-
	OTau =.. [ C | OArgs ],
	map_lub_from_originals(OArgs,Args),
	Tau =.. [ C | Args ].
map_lub_from_originals([],[]).
map_lub_from_originals([Origs|Rest],[LUB|OtherLUB]) :-
	tclp__multiple_vLUB(Origs,LUB),
	map_lub_from_originals(Rest,OtherLUB).

%% replace_by_glb(OTau,Tau)
%% OTau is the type constructor of Tau with the list of original variable
%% under the arguments of Tau as its arguments
replace_by_glb(OTau,Tau) :-
	OTau =.. [ C | OArgs ],
	map_glb_from_originals(OArgs,Args),
	Tau =.. [ C | Args ].
map_glb_from_originals([],[]).
map_glb_from_originals([Origs|Rest],[GLB|OtherGLB]) :-
	tclp__multiple_vGLB(Origs,GLB),
	map_glb_from_originals(Rest,OtherGLB).

%%%%%%%%%
%% misc operations

%% adds the constraints A :< B for all A in the first argument and all B in
%% the second
link([],_).
link([A|Rest],Bs) :-
	link2(A,Bs), link(Rest,Bs).
link2(_,[]).
link2(A,[B|Rest]) :-
	A :< B, link2(A,Rest).
